# Copyright (c) 2013 Dirk Jagdmann <doj@cubic.org>

# This software is provided 'as-is', without any express or implied
# warranty. In no event will the authors be held liable for any damages
# arising from the use of this software.

# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:

#     1. The origin of this software must not be misrepresented; you
#        must not claim that you wrote the original software. If you use
#        this software in a product, an acknowledgment in the product
#        documentation would be appreciated but is not required.

#     2. Altered source versions must be plainly marked as such, and
#        must not be misrepresented as being the original software.

#     3. This notice may not be removed or altered from any source
#        distribution.

# $Id: PayPal.pm 7554 2013-12-22 23:37:24Z doj $

package PayPal;

use Carp;
use Time::Local;
use Term::ReadKey;
use JSON;
use configuration;

sub ts2str
{
    my ($ts) = @_;
    return unless $ts > 0;
    my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = gmtime($ts);
    return sprintf('%i-%02i-%02iT%02i:%02i:%02iZ', $year+1900, $mon+1, $mday, $hour, $min, $sec);
}

sub str2ts
{
    if ($_[0] =~ /(\d{4})-(\d\d)-(\d\d)T(\d\d):(\d\d):(\d\d)Z/)
    {
	return timegm($6,$5,$4,$3,$2 - 1,$1);
    }
    return 0;
}

sub str2local_date
{
    my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime(str2ts($_[0]));
    return sprintf('%02i/%02i/%04i', $mon+1, $mday, $year+1900);
}

sub iif
{
    my $iif = shift;

    unless (Quickbooks::IIF::init()) {
	return;
    }

    foreach my $txn (@_)
    {
	my $h = PayPal::DB::get_txn($txn);
	return unless ref($h) eq 'HASH';

	# lookup other income account name
	my $title = $h->{'Title'};
	my $quickbooks_account_name;
	my $rid;

	if ($h->{Type} eq 'Refund')
	{
	    # find original transaction ID
	    $rid = $h->{ParentTransactionID};
	    unless ($rid) { carp "ParentTransactionID not defined for Refund $txn"; return }
	    my $r = PayPal::DB::get_txn($rid);
	    if (ref($r) eq 'HASH')
	    {
		# find original item title, to look up original Other Income account name
		my $rit = $r->{'Title'};
		if ($configuration::handleUnknownItemTitle == configuration::DIE_ON_UNKNOWN_ITEM_TITLE)
		{
		    unless (exists $configuration::ItemTitle2Income{$rit}) { carp "no mapping defined for PayPal item title '$rit'"; return }
		}
		$quickbooks_account_name = $configuration::ItemTitle2Income{$rit} if exists $configuration::ItemTitle2Income{$rit};
	    }
	}
	elsif ($h->{Type} eq "Withdraw Funds to a Bank Account")
	{
	    $quickbooks_account_name = $configuration::checking_account_name;
	}

	if ($quickbooks_account_name)
	{
	    # do nothing
	}
	elsif (exists($configuration::ItemTitle2Income{$title}))
	{
	    $quickbooks_account_name = $configuration::ItemTitle2Income{$title};
	}
	else
	{
	    if ($force || $configuration::handleUnknownItemTitle == configuration::USE_OTHER_INCOME_ON_UNKNOWN_ITEM_TITLE)
	    {
		$quickbooks_account_name = $configuration::other_income;
	    }
	    elsif ($configuration::handleUnknownItemTitle == configuration::ASK_ON_UNKNOWN_ITEM_TITLE)
	    {
		$quickbooks_account_name = interactiveQuickbooksAccountSelect($h);
		unless($quickbooks_account_name) {
		    carp "could not determine Quickbooks account name";
		    return
		}
	    }
	    else
	    {
		unless ($configuration::handleUnknownItemTitle == configuration::DIE_ON_UNKNOWN_ITEM_TITLE) {
		    carp "\$configuration::handleUnknownItemTitle has unexpected value: $configuration::handleUnknownItemTitle";
		    return
		}
		carp "$txn unmapped item title: '$title', transaction type '$h->{Type}'";
		return
	    }
	}

	unless ($quickbooks_account_name) {
	    carp "$txn could not determine Quickbooks account name";
	    return
	}

	# add a description to the title
	if ($h->{Type} eq 'Refund')
	{
	    $title .= " refund $rid";
	}
	else
	{
	    $title .= " $txn";
	}

	$h->{PayPalAccount} = $configuration::paypal_account_name;
	$h->{QuickbooksAccount} = $quickbooks_account_name;
	$h->{Title} = $title;
	unless (Quickbooks::IIF::addrecord($h)) {
	    print STDERR to_json($h, { utf8 => 1, pretty => 1 } );
	    carp "could not add IIF record";
	    return
	}
	print STDERR "converted ${txn}: $title -> $quickbooks_account_name $h->{Name} $h->{NetAmount}\n";
    }

    # create output IIF file
    my $res = Quickbooks::IIF::save($iif);
    return $#_ + 1 if $res > 0;
    return $res;
}

sub choose_account_name
{
    print "\n";

    # put all account names into @a
    my @a = ($configuration::checking_account_name, $configuration::other_income);
    while(my ($k,$v) = each(%configuration::ItemTitle2Income)) {
	push(@a, $v);
    }
    @a = sort @a;

    # print @a
    for(my $i = 0; $i <= $#a; ++$i) {
	my $is = $i + 1;
	print "${is}: $a[$i]\n";
    }
    local $| = 1; # auto flush STDOUT
    print "\n> (number, new name, 'd' details, 'f' full details, 's' summary): ";

    # select an account
    my $i = ReadLine(0);
    chomp $i;
    unless (length($i)) {
	print STDERR "no Quickbooks account choosen\n";
	return
    }

    if ($i eq 's') {
	return -1;
    }
    if ($i eq 'd') {
	return -2;
    }
    if ($i eq 'f') {
	return -3;
    }

    # if only digits were entered, lookup account from list.
    if ($i =~ /^\d+$/) {
	$i += 0; # make integer out of string
	if ($i == 0) {
	    print STDERR "zero not valid";
	    return
	}
	--$i; # make zero based index
	if ($i > $#a) {
	    print STDERR "invalid number";
	    return
	}
	return $a[$i];
    }

    # use input as account name directly
    if ($i =~ /^\s+$/) {
	print STDERR "empty account name";
	return
    }
    return $i;
}

sub interactiveQuickbooksAccountSelect
{
    my ($h) = @_;
    unless (ref($h) eq 'HASH') {
	carp "invalid parameter for interactiveQuickbooksAccountSelect()";
	return
    }

    my $res = -1;
    while ($res < 0) {
	print "\n\nSelect a Quickbooks account name for the following transaction:\n";
	if ($res == -1) {
	    print<<"EOT";
Type:  $h->{Type} $h->{TransactionType} $h->{PII_Name}
Gross: $h->{GrossAmount}
Name:  $h->{PayerDisplayName} $h->{Payer}
EOT
	} elsif ($res == -2) {
	    foreach my $k (sort keys %$h) {
		# only print if key and value are not empty
		next unless length($k) > 0;
		my $v = $h->{$k};
		next unless length($v) > 0;
		print "${k}:\t$v\n";
	    }
	} elsif ($res == -3) {
	    print to_json($h, {utf8 => 1, pretty => 1});
	} else {
	    die "invalid \$res $res";
	}
	$res = choose_account_name();
    }

    return $res;
}

1;
